import numpy as np
import re
from itertools import islice


class ReadIFile:
	def __init__(self,f):
		self.infile = f
	def get_data(self):
		file = open(self.infile,"r")
		regexStart = re.compile('\$')
		regexEnd = re.compile('\*')
		P1atom_list = []
		P2atom_list = []
		xtal_list = []
		p1vol_list = []
		p2vol_list = []
		acoustic_list  =[]
		for line in file:
			if(regexStart.search(line) != None):
				Label = (''.join(e for e in line if e.isalnum())).lower()
			elif(regexEnd.search(line) != None):
				continue
			elif(line == "\n"):
				Label = 'Blank'
			else:
				if Label == 'P1Atoms'.lower():
					P1atom_list = np.append(P1atom_list,line)
				elif Label == 'P2Atoms'.lower():
					P2atom_list = np.append(P2atom_list,line)
				elif Label == 'Crystals'.lower():
					xtal_list = np.append(xtal_list,line)
				elif Label == 'Acoustics'.lower():
					acoustic_list = np.append(acoustic_list,line)
				elif Label == 'P1Volumes'.lower():
					p1vol_list = np.append(p1vol_list,line)
				elif Label == 'P2Volumes'.lower():
					p2vol_list = np.append(p2vol_list,line)
				elif Label == 'Pmax'.lower():
					Pmax = float(line)
				elif Label == 'Blank':
					continue
		file.close()
		def AtomIDS(atom_list):
			atom_array = []
			for x in range(np.size(atom_list)):
				test = atom_list[x].split()
				if test[0] == 'C':
					num_c = int(test[1])
				if test[0] == 'H':			
					num_h = int(test[1])
				if test[0] == 'O':
					num_o = int(test[1])
				if test[0] == 'N':
					num_n = int(test[1])
				if test[0] == 'S':
					num_s = int(test[1])
			atom_array = [num_c, num_h, num_o, num_n, num_s]
			return atom_array
		
		def AcousticJobID(acoustic_list):
			for x in range(np.size(acoustic_list)):
				test = acoustic_list[x].split()
				if test[0] == 'EC':
					EC_ID = test[1]
				if test[0] == 'Christoffel':
					Chris_ID = test[1]
			acoustic_job = [EC_ID, Chris_ID]
			return acoustic_job
		
		P1atom_array = AtomIDS(P1atom_list)
		if len(P2atom_list) == 0:
			P2atom_array = 0
		else:
			P2atom_array = AtomIDS(P2atom_list)	
		acoustic_job = AcousticJobID(acoustic_list)
	
		return P1atom_array, P2atom_array, xtal_list, acoustic_job, p1vol_list, p2vol_list, Pmax

class ReadYaml:
	def __init__(self, f):
		self.infile = f
	def get_lattice(self):
		file = open(self.infile, "r")
		LattTag = 'lattice:'
		for line in file:
			if line.rstrip()== LattTag:
				latt = (list(islice(file,0,3)))
				break
		lattice = []
		for i in range(np.size(latt)):
			tmp = (latt[i].replace('- [','').replace(',','').replace(']','')).split()
			tmp = np.double(tmp[0:3])
			lattice = np.append(lattice,tmp)
		file.close()
		return lattice

	def get_SS(self):
		file = open(self.infile, "r")
		FreqTag = 'frequency:'
		KptTag = '- q-position:'
		Frequencies = []
		Kpoints = []
		for line in file:
			line = list(line)
			line = ''.join(line)
			try:
				(line.index(FreqTag))
			except ValueError:
				pass
			else: 
				line = (line.replace(FreqTag,""))
				Frequencies.append(line)
			try:
				(line.index(KptTag))
			except ValueError:
				pass
			else:
				line = (line.replace(KptTag,""))
				line =line.replace('[','').replace(',','').replace(']','')
				line = line.split(' ')
				line = [ele for ele in line if ele][0:3]
				Kpoints.append(line)
		Kpoints = np.array(Kpoints)
		Kpoints = np.double(Kpoints)
		Frequencies = np.array(Frequencies)
		Frequencies = ''.join(Frequencies)
		Frequencies = np.fromstring(Frequencies, dtype = float,sep = '\n')
		file.close()
		return Frequencies, Kpoints
	
	def get_UC(self):
		file = open(self.infile, "r")
		FreqTag = 'frequency:'
		AtomTag = '# atom'
		Frequencies = []
		Eigenvectors = []
		for line in file:
			line = list(line)
			line = ''.join(line)
			try:
				(line.index(FreqTag))
			except ValueError:
				pass
			else: 
				line = (line.replace(FreqTag,""))
				Frequencies.append(line)
			try:
				(line.index(AtomTag))
			except ValueError:
				pass
			else:
				vector = []
				evec = list(islice(file,3))
				evec = ''.join(evec).split('\n')
				for x in evec:
					tmp = x.replace('- [','').replace(',','').replace(']','')
					tmp = tmp.split(' ')
					tmp = [ele for ele in tmp if ele]
					if not tmp:
						continue
					else:
						evec2 = tmp[0]
					vector.append(evec2)
				test = np.array(vector)
				Eigenvectors = np.append(Eigenvectors, test)
		Eigenvectors = Eigenvectors.reshape([-1,3])
		Eigenvectors = np.double(Eigenvectors)
		Frequencies = np.array(Frequencies)
		Frequencies = ''.join(Frequencies)
		Frequencies = np.fromstring(Frequencies, dtype = float,sep = '\n')
		file.close()
		return Frequencies, Eigenvectors
	
class ReadElastic:
	def __init__(self, t):
		self.tensor = t
	def get_tensor(self):
		file = open(self.tensor,"r")
		tens_array = []
		for line in file:
			line = line.split()
			tens_array.append(line)
		tens_array = np.array(tens_array)
		tens_array = tens_array.astype(float)
		file.close()
		return tens_array

class ReadLattice:
	def __init__(self,l):
		self.lattice = l
	def get_vecs(self):
		file = open(self.lattice)
		file = list(file)
		vecs = []
		for line in file:
			line = line.split()
			vecs.append(line)
		vecs = np.array(vecs)
		vecs = vecs.astype(float)
		file.close()
		return vecs
class Quasiharmonic:
	def __init__(self, v):
		self.evcurve = v
	def get_evcurve(self):
		file = open(self.evcurve,"r")
		ev_data = list(file)
		ev_data = [x for x in ev_data if "#" not in x]
		ev_data = ''.join(ev_data)
		ev_data = np.fromstring(ev_data, dtype = float, sep = '\n')
		ev_data = np.reshape(ev_data,(-1,2))
		file.close()
		return ev_data




from . import ReadData as data
from . import JobControl 
import numpy as np
import sys

def RunJob():


	print('Modematch v0.0.5')
	sys.setrecursionlimit(3000) ## DEFAULT RECURSION LIM HIT WITH STABLE MARRIAGE MATCHING ALG


	job_description = data.ReadIFile('infile') #Chnge this when im done(?)
	job_params = job_description.get_data()

	P1atom_IDs = job_params[0]
	P2atom_IDs = job_params[1]
	struct_IDs = job_params[2]
	acoustic_ID = job_params[3]
	p1freq_vols = job_params[4]
	p2freq_vols = job_params[5]
	Pmax = job_params[6]

	Press = np.arange(0,Pmax+0.01,0.01)

	if P2atom_IDs == 0:
		PhaseTrans = False
		print('Only 1 structure, no phase trans...')
		natoms = sum(P1atom_IDs)
		#dim = 3 * natoms
		poly1 = struct_IDs[0].rstrip()

	else:
		PhaseTrans = True
		print('Multiple structures. Predicting Phase Transition...')
		P1atoms = sum(P1atom_IDs)
		P2atoms = sum(P2atom_IDs)
		#P1dim = 3 * P1atoms
		#P2dim = 3 * P2atoms
		poly1 = struct_IDs[0].rstrip()
		poly2 = struct_IDs[1].rstrip()

	#Number of Fvibs to calculate for each structure
	P1Count = int(len(p1freq_vols))
	P2Count = int(len(p2freq_vols))

	if PhaseTrans == False:
		P1Control = JobControl.Control(poly1, P1Count, P1atom_IDs)
		P1Control.MakePaths()

		if acoustic_ID[0].lower() == 'true' and acoustic_ID[1].lower() == 'true':
			AllConstants = P1Control.SolveECs()
			NewAcoustics = P1Control.Dispersion(AllConstants)
			ShiftedFreqs = P1Control.ShiftPlusECs(NewAcoustics)

		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'false': ##No EC acoustics at all
			ShiftedFreqs = P1Control.Modematch()[0]

		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'true': ##User input EC's
			AllConstants = P1Control.ReadECs()
			NewAcoustics = P1Control.Dispersion(AllConstants)
			ShiftedFreqs = P1Control.ShiftPlusECs(NewAcoustics)

		#Now we have the shifted freqs...
		if P1Count == 1:
			print("One structure, no QHA")
			import GetDOS as dos
			import pandas as pd
			[Fvib, Hvib, Svib,T] = dos.EvaluateDOS(ShiftedFreqs,natoms)
			ThermoData = pd.DataFrame({'Temperature':np.array(T),
										'Enthalpy': np.array(Hvib),
										'Entropy': np.array(Svib),
										'Helmholtz': np.array(Fvib)})
		else:
			ThermoData = P1Control.QuasiHarmonic(ShiftedFreqs,p1freq_vols,Press)

		filename = poly1 + '.csv'
		ThermoData.to_csv(filename,index=False)
		print('JOB SUCCESSFUL!!')

	else:
		##FIRST STRUCTURE
		P1Control = JobControl.Control(poly1, P1Count, P1atom_IDs)
		P1Control.MakePaths()
		if acoustic_ID[0].lower() == 'true' and acoustic_ID[1].lower() == 'true':
			AllConstants = P1Control.SolveECs()
			NewAcoustics = P1Control.Dispersion(AllConstants)
			ShiftedFreqs = P1Control.ShiftPlusECs(NewAcoustics)

		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'false': ##No EC acoustics at all
			ShiftedFreqs = P1Control.Modematch()[0]

		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'true': ##User input EC's
			AllConstants = P1Control.ReadECs()
			NewAcoustics = P1Control.Dispersion(AllConstants)
			ShiftedFreqs = P1Control.ShiftPlusECs(NewAcoustics)
		#Now we have the shifted freqs...
		if P1Count == 1:
			P1CheckQHA = False
			print("One structure, no QHA")
			import GetDOS as dos
			import pandas as pd
			[Fvib, Hvib, Svib,T] = dos.EvaluateDOS(ShiftedFreqs,P1atoms)
			P1ThermoData = pd.DataFrame({'Temperature':np.array(T),
										'Enthalpy': np.array(Hvib),
										'Entropy': np.array(Svib),
										'Helmholtz': np.array(Fvib)})
		else:
			P1CheckQHA = True
			P1ThermoData = P1Control.QuasiHarmonic(ShiftedFreqs,p1freq_vols,Press)
		filename = poly1 + '.csv'
		P1ThermoData.to_csv(filename,index=False)
		print('DONE WITH ', poly1)

		##SECOND STRUCTURE
		P2Control = JobControl.Control(poly2, P2Count, P2atom_IDs)
		P2Control.MakePaths()

		#Check to see if we're calculating new acoustics / ECs
		if acoustic_ID[0].lower() == 'true' and acoustic_ID[1].lower() == 'true':
			AllConstants = P2Control.SolveECs()
			NewAcoustics = P2Control.Dispersion(AllConstants)
			ShiftedFreqs = P2Control.ShiftPlusECs(NewAcoustics)

		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'false': ##No EC acoustics at all
			ShiftedFreqs = P2Control.Modematch()[0]


		elif acoustic_ID[0].lower() == 'false' and acoustic_ID[1].lower() == 'true': ##User input EC's
			AllConstants = P2Control.ReadECs()
			NewAcoustics = P2Control.Dispersion(AllConstants)
			ShiftedFreqs = P2Control.ShiftPlusECs(NewAcoustics)

		#Now we have the shifted freqs...
		if P2Count == 1:
			P2CheckQHA = False
			print("One structure, no QHA")
			import GetDOS as dos
			import pandas as pd
			[Fvib, Hvib, Svib,T] = dos.EvaluateDOS(ShiftedFreqs,P2atoms)
			P2ThermoData = pd.DataFrame({'Temperature':np.array(T),
										'Enthalpy': np.array(Hvib),
										'Entropy': np.array(Svib),
										'Helmholtz': np.array(Fvib)})
		else:
			P2CheckQHA = True
			P2ThermoData = P2Control.QuasiHarmonic(ShiftedFreqs,p2freq_vols,Press)
		filename = poly2 + '.csv'
		P2ThermoData.to_csv(filename,index=False)
		print('DONE WITH ', poly2)


		if P1CheckQHA and P2CheckQHA == True:
			PhaseTransData = P1Control.PhaseTransControl(P1ThermoData,P2ThermoData,Press)
			filename = 'Phase_Transition'
			PhaseTransData.to_csv(filename,index=False)

		print('JOB SUCCESSFUL!!')



import numpy as np
from . import GetDOS as dos
from scipy import optimize
import pandas as pd

def QHA(AllFreqs,FreqVols,ev_curve,count,natoms,Press):
	print('STARTING QHA')
	Na = 6.0221367e23
	natoms = sum(natoms)
	dim = 3 * natoms
	AllFreqs = AllFreqs.transpose()

#Curve Data
	Volumes = ev_curve[:,0]
	Energies = ev_curve[:,1]
	Vmin = min(Volumes)
	Vmax = max(Volumes)

	TArray = []	
	OptVolArray = []
	OptGibbsArray = []
	OptFvibArray = [] 
	OptSvibArray = []
	OptHvibArray = []
	OptElArray = []

	FvibAll = []
	HvibAll = []
	SvibAll = []
	TAll = []

	for i in range(np.size(AllFreqs,1)):
		freqset = np.array(AllFreqs[:,i]).reshape(-1,dim)
		[Fvib, Hvib, Svib,T] = dos.EvaluateDOS(freqset,natoms)
		print('Done with frequency set', i)
		FvibAll = np.append(FvibAll, Fvib)
		HvibAll = np.append(HvibAll, Hvib)
		SvibAll = np.append(SvibAll, Svib)
		TAll = np.append(TAll, T)
	

	ThermoDim = np.int(np.size(FvibAll)/count)
	FvibAll = np.array(FvibAll).reshape(-1,ThermoDim)
	HvibAll = np.array(HvibAll).reshape(-1,ThermoDim)
	SvibAll = np.array(SvibAll).reshape(-1,ThermoDim)
	TAll = np.array(TAll).reshape(-1,ThermoDim)
	np.savetxt("Fvibs.csv",FvibAll,delimiter=",")
	print('Starting pressure scan...')

	for P in Press:
	#Loop over all T's evaluatied in Thermo (Columns in FvibAll)
		for i in range(ThermoDim):
			F_polyfit = np.polyfit(FreqVols, FvibAll[:,i], 2)
			H_polyfit = np.polyfit(FreqVols, HvibAll[:,i], 2)
			S_polyfit = np.polyfit(FreqVols, SvibAll[:,i], 2)
			PolyFvibFunction = np.polyval(F_polyfit,Volumes)
			PV = P * Volumes * 1.0e-24 * Na
			Gibbs = PolyFvibFunction + Energies + PV
			G_polyfit = np.polyfit(Volumes,Gibbs,4)

		# EP CHECKER Don't allow the minimum to be on the endpoints
			MinIndex = np.argmin(Gibbs)
			if Volumes[MinIndex] == Volumes[-1]:
				MinIndex = MinIndex - 1
			elif Volumes[MinIndex] == Volumes[0]:
				MinIndex = MinIndex + 1

		#Define initial Gibbs curve \\ G(V) 
			def PolyGFxn(V):
				return np.polyval(G_polyfit,V)
		
			##Get initial V for minimumization
			VGrid = np.arange(Vmin,Vmax,0.01)
			T_fit = np.polyval(G_polyfit,VGrid)
			initV = VGrid[np.argmin(T_fit)]

			opt_params = optimize.minimize(PolyGFxn,initV) #Provides initial Vmin and Gmin for Murnaghan Fit
			minimum = opt_params.x[0]
			minGibbs = opt_params.fun

		#Murnaghan EOS \\ Check parentheses
			def Murnaghan(x,c0,c1):
				return x[:,2] + c0*x[:,1] * (1 / (c1*(c1 - 1)) * np.power((x[:,0]/x[:,1]),1-c1) + 1 / c1 * (x[:,0]/x[:,1]) - 1 / (c1-1))

		#Figure out how to combine these fxns... it was an indexing issue
			def DoubleMurn(x,Vmin,Gmin,comp,exp): 
	
				def Murnaghan(x,c0,c1):
					return  x[2] + c0*x[1] * (1 / (c1*(c1 - 1)) * np.power((x[0]/x[1]),1-c1) + 1 / c1 * (x[0]/x[1]) - 1 / (c1-1))
				
				Test = np.zeros((1,3))
				Test[:,0] = x
				Test[:,1] = Vmin
				Test[:,2] = Gmin
				out = []

				if x > Vmin:
					ans = Murnaghan(Test[0,:],exp[0],exp[1])
					out = np.append(out,ans)
				else:
					ans = Murnaghan(Test[0,:],comp[0],comp[1])
					out = np.append(out,ans)
				return out

		#Define Compression and Expansion branches for Double Murn
			if P < 0.75:
				VolComp = Volumes[MinIndex - 1:-1]
				VolExp = Volumes[0:MinIndex  + 6]
				GibbsComp = Gibbs[MinIndex - 1:-1]
				GibbsExp = Gibbs[0:MinIndex  + 6]

			else:
				VolComp = Volumes[MinIndex - 3:-1]
				VolExp = Volumes[0:MinIndex  + 1]
				GibbsComp = Gibbs[MinIndex - 3:-1]
				GibbsExp = Gibbs[0:MinIndex  + 1]
		
		#Sets up weights for the nonlinear fit applied to Murn (Unnecessary?) BREAKS CODE; OVERWRITES VOLS
		#Wcomp = VolComp
		#for i in range(len(VolComp)):
		#	if abs(Wcomp[i] - Gmin_Volume) == 0:
		#		Wcomp[i] = 1.0
		#	elif abs(Wcomp[i] - Gmin_Volume) < 50:
		#		Wcomp[i] = 0.25
			#else:
				#Wcomp[i]
		#Wexp = VolExp
		#for i in range(len(VolExp)):
		#	if abs(Wexp[i] - Gmin_Volume) == 0:
		#		Wexp[i] = 1.0
		#	elif abs(Wexp[i] - Gmin_Volume) < 50:
		#		Wexp[i] = 0.25
		#	else:
		#		Wexp[i] = 0.1

		#Get optimized parameters for Compression and Expansion branches
			a = [5.0,8.0]
			Comp = np.zeros((len(VolComp),3))
			Comp[:,0] = VolComp
			Comp[:,1] = minimum
			Comp[:,2] = minGibbs
			Exp = np.zeros((len(VolExp),3))
			Exp[:,0] = VolExp
			Exp[:,1] = minimum
			Exp[:,2] = minGibbs

		#Optimize parameters
			Exp_coeffs = optimize.curve_fit(Murnaghan,Exp, GibbsExp,a)[0]
			Comp_coeffs = optimize.curve_fit(Murnaghan,Comp,GibbsComp,a)[0]

			#OptimalVolume = optimize.brent(DoubleMurn,args=(minimum,minGibbs,Comp_coeffs,Exp_coeffs),brack=(Vmin,Vmax))
			#OptimalGibbs = DoubleMurn(OptimalVolume,minimum,minGibbs,Comp_coeffs,Exp_coeffs)
			OptimalParams = optimize.minimize(DoubleMurn,minimum,args=(minimum,minGibbs,Comp_coeffs,Exp_coeffs))
			OptimalVolume = OptimalParams.x[0]
			OptimalGibbs = OptimalParams.fun
			OptimalFvib = np.polyval(F_polyfit,OptimalVolume)
			OptimalSvib = np.polyval(S_polyfit,OptimalVolume)
			OptimalHvib = np.polyval(H_polyfit,OptimalVolume)
			#OptimalPV = P * OptimalVolume * 1E-24 * Na
			OptimalEl = OptimalGibbs - OptimalFvib # - OptimalPV

			TArray = np.append(TArray,TAll[1,i])
			OptVolArray = np.append(OptVolArray, OptimalVolume)
			OptGibbsArray = np.append(OptGibbsArray, OptimalGibbs)
			OptFvibArray = np.append(OptFvibArray, OptimalFvib)
			OptSvibArray = np.append(OptSvibArray, OptimalSvib)
			OptHvibArray = np.append(OptHvibArray, OptimalHvib)
			OptElArray = np.append(OptElArray, OptimalEl)

		AllData = pd.DataFrame({'Temperature': np.array(TArray),
							'Volume': np.array(OptVolArray),
							'Gibbs': np.array(OptGibbsArray),
							'Helmholtz': np.array(OptFvibArray),
							'Entropy': np.array(OptSvibArray),
							'Enthalpy': np.array(OptHvibArray),
							'Electronic_E' : np.array(OptElArray)})
	return AllData	
	
def PhaseTrans(AllData1,AllData2,Press):
	
	NumPress = np.size(Press)
	NumTemps = int(np.size(AllData1['Temperature'][:]) / np.size(Press))

	T_Trans_Array = []
	S_Trans_Array = []
	H_Trans_Array = []

	for i in range(NumPress):

		GibbsDiff = AllData2['Gibbs'][i * NumTemps: (i + 1) * NumTemps] - AllData1['Gibbs'][i * NumTemps: (i + 1) * NumTemps]
		Temps = AllData2['Temperature'][i * NumTemps: (i + 1) * NumTemps]

		T_fit = np.polyfit(Temps,GibbsDiff,3)
		TempTrans = np.roots(T_fit)[2]

		Enthalpy1 = AllData1['Electronic_E'][i * NumTemps: (i + 1) * NumTemps] + AllData1['Enthalpy'][i * NumTemps: (i + 1) * NumTemps]
		Poly1_Sfit = np.polyfit(Temps, AllData1['Entropy'][i * NumTemps: (i + 1) * NumTemps], 2)
		Poly1_Hfit = np.polyfit(Temps, Enthalpy1, 3)

		TransitionEntropy1 = np.polyval(Poly1_Sfit,TempTrans)
		TransitionEnthalpy1 = np.polyval(Poly1_Hfit,TempTrans)

		Enthalpy2 = AllData2['Electronic_E'][i * NumTemps: (i + 1) * NumTemps] + AllData2['Enthalpy'][i * NumTemps: (i + 1) * NumTemps]
		Poly2_Sfit = np.polyfit(Temps, AllData2['Entropy'][i * NumTemps: (i + 1) * NumTemps], 2)
		Poly2_Hfit = np.polyfit(Temps, Enthalpy2,3)

		TransitionEntropy2 = np.polyval(Poly2_Sfit,TempTrans)
		TransitionEnthalpy2 = np.polyval(Poly2_Hfit,TempTrans)
	
		dS = TransitionEntropy2 - TransitionEntropy1
		dH = TransitionEnthalpy2 - TransitionEnthalpy1
		

		T_Trans_Array = np.append(T_Trans_Array, TempTrans)
		S_Trans_Array = np.append(S_Trans_Array, dS)
		H_Trans_Array = np.append(H_Trans_Array, dH)

	Phase_Trans_Data = pd.DataFrame({'T_Trans': np.array(T_Trans_Array),
								 'H_Trans': np.array(H_Trans_Array),
								 'S_Trans': np.array(S_Trans_Array),
								 'Pressure': Press })

	print(Phase_Trans_Data)
	return Phase_Trans_Data



